import time
import multiprocessing
import multiprocessing.pool

import os
try:
    import Queue as queue
except ImportError:
    import queue
import multiprocessing
import multiprocessing.pool
from datetime import datetime

from paegan.transport.particles.particle import Particle
from paegan.transport.controllers import BaseModelController
from paegan.transport.forcers import CachingForcer
from paegan.transport.parallel_manager import CachingDataController, Consumer
from paegan.utils.asarandom import AsaRandom
from paegan.logger import logger
import paegan.transport.export as ex


class CachingModelController(BaseModelController):

    def __init__(self, **kwargs):
        super(CachingModelController, self).__init__(**kwargs)
        self.time_chunk  = kwargs.get('time_chunk', 10)
        self.horiz_chunk = kwargs.get('horiz_chunk', 5)

    def total_task_count(self):
        # Add the CachingDataController to the number of tasks
        return self.total_particle_count() + 1

    def start_tasks(self, **kwargs):
        try:
            logger.info('Starting CachingDataController')

            # Add data controller to the queue first so that it
            # can get the initial data and is not blocked
            data_controller = CachingDataController(self.hydrodataset, self.common_variables, self.n_run, self.get_data, self.write_lock, self.has_write_lock, self.read_lock, self.read_count,
                                                    self.time_chunk, self.horiz_chunk, self.times, self.start, self.point_get, self.reference_location, cache_path=self.cache_path)
            self.tasks.put(data_controller)
            # Create CachingDataController worker
            self.data_controller_process = Consumer(self.tasks, self.results, self.n_run, self.nproc_lock, self.active, self.get_data, name="CachingDataController")
            self.data_controller_process.start()

            logger.info('Adding %i particles as tasks' % self.total_particle_count())

            for part in self.particles:
                forcer = CachingForcer(self.cache_path,
                                       particle=part,
                                       common_variables=self.common_variables,
                                       timevar=self.timevar,
                                       times=self.times,
                                       start_time=self.start,
                                       models=self._models,
                                       release_location_centroid=self.reference_location.point,
                                       usebathy=self._use_bathymetry,
                                       useshore=self._use_shoreline,
                                       usesurface=self._use_seasurface,
                                       reverse_distance=self.reverse_distance,
                                       bathy_path=self.bathy_path,
                                       shoreline_path=self.shoreline_path,
                                       shoreline_feature=self.shoreline_feature,
                                       time_method=self.time_method,
                                       shoreline_index_buffer=self.shoreline_index_buffer,
                                       get_data=self.get_data,
                                       read_lock=self.read_lock,
                                       has_read_lock=self.has_read_lock,
                                       read_count=self.read_count,
                                       point_get=self.point_get,
                                       data_request_lock=self.data_request_lock,
                                       has_data_request_lock=self.has_data_request_lock
                                      )
                self.tasks.put(forcer)

            # Create workers for the particles.
            self.procs = [Consumer(self.tasks, self.results, self.n_run, self.nproc_lock, self.active, self.get_data, name="CachingForcer-%d" % i)
                          for i in range(self.nproc - 1) ]
            logger.progress((5, 'Running model'))
            for w in self.procs:
                w.start()
                logger.info('Started %s' % w.name)

            return True

        except Exception:
            logger.exception("Something didn't start correctly!")
            return False

    def __str__(self):
        return "*** CachingModelController ***"

    def setup_run(self, hydrodataset, **kwargs):

        super(CachingModelController, self).setup_run(hydrodataset, **kwargs)

        # Get the number of cores (may take some tuning) and create that
        # many workers then pass particles into the queue for the workers
        self.mgr = multiprocessing.Manager()

        # This tracks if the system is 'alive'.  Most looping whiles will check this
        # and break out if it is False.  This is True until something goes very wrong.
        self.active = self.mgr.Value('bool', True)

        # Either spin up the number of cores, or the number of tasks
        self.nproc = min(multiprocessing.cpu_count() - 1, self.total_task_count())

        # Number of tasks that we need to run.  This is decremented everytime something exits.
        self.n_run = self.mgr.Value('int', self.total_task_count())

        # The lock that controls access to the 'n_run' variable
        self.nproc_lock = self.mgr.Lock()

        # Create the task queue for all of the particles and the CachingDataController
        self.tasks = multiprocessing.JoinableQueue(self.total_task_count())

        # Create the result queue for all of the particles and the CachingDataController
        self.results = self.mgr.Queue(self.total_task_count())

        # Should we remove the cache file at the end of the run?
        self.remove_cache        = kwargs.get("remove_cache", False)
        self.cache_path          = kwargs.get("cache_path", None)

        # Create a temp file for the cache if nothing was passed in
        if self.cache_path is None:
            default_cache_dir = os.path.join(os.path.dirname(__file__), "_cache")
            temp_name = AsaRandom.filename(prefix=str(datetime.now().microsecond), suffix=".nc")
            self.cache_path = os.path.join(default_cache_dir, temp_name)

        # Be sure the cache directory exists
        if not os.path.exists(os.path.dirname(self.cache_path)):
            logger.info("Creating cache directory: %s" % self.cache_path)
            os.makedirs(os.path.dirname(self.cache_path))

        # Create the shared state objects

        # Particles use this to tell the Data Controller to "get_data".
        # The CachingDataController sets this to False when it is done writing to the cache file.
        # Particles will wait for this to be False before reading from the cache file.
        # If we are caching, this starts as True so the Particles don't take off.  If we
        # are not caching, this is False so the Particles can start immediatly.
        self.get_data = self.mgr.Value('bool', True)
        # Particles use this to tell the DataContoller which indices to 'get_data' for
        self.point_get = self.mgr.Value('list', [0, 0, 0])

        # This locks access to the 'has_data_request_lock' value
        self.data_request_lock = self.mgr.Lock()
        # This tracks which Particle PID is asking the CachingDataController for data
        self.has_data_request_lock = self.mgr.Value('int', -1)

        # The lock that controls access to modifying 'has_read_lock' and 'read_count'
        self.read_lock = self.mgr.Lock()
        # List of Particle PIDs that are reading from the cache
        self.has_read_lock = self.mgr.list()
        # The number of Particles that are reading from the cache
        self.read_count = self.mgr.Value('int', 0)

        # When something is writing to the cache file
        self.write_lock = self.mgr.Lock()
        # PID of process with lock
        self.has_write_lock = self.mgr.Value('int', -1)

    def listen_for_results(self, output_h5_file, total_particles):
        try:
            # Get results back from queue, test for failed particles
            return_particles = []
            retrieved = 0.
            self.error_code = 0

            logger.info("Waiting for %i particle results" % total_particles)
            while retrieved < self.total_task_count():  # One for the CachingDataController

                logger.info("looping in listen_for_results")

                try:
                    # Returns a tuple of code, result
                    code, tempres = self.results.get(timeout=240)
                except queue.Empty:
                    # Poll the active processes to make sure they are all alive and then continue with loop
                    if not self.data_controller_process.is_alive() and self.data_controller_process.exitcode != 0:
                        # Data controller is zombied, kill off other processes.
                        self.get_data.value is False
                        self.results.put((-2, "CachingDataController"))

                    new_procs = []
                    old_procs = []
                    for p in self.procs:
                        if not p.is_alive() and p.exitcode != 0:
                            # Do what the Consumer would do if something finished.
                            # Add something to results queue
                            self.results.put((-3, "ZombieParticle"))
                            # Decrement nproc (CachingDataController exits when this is 0)
                            with self.nproc_lock:
                                self.n_run.value = self.n_run.value - 1

                            # Remove task from queue (so they can be joined later on)
                            self.tasks.task_done()

                            # Start a new Consumer.  It will exit if there are no tasks available.
                            np = Consumer(self.tasks, self.results, self.n_run, self.nproc_lock, self.active, self.get_data, name=p.name)
                            new_procs.append(np)
                            old_procs.append(p)

                            # Release any locks the PID had
                            if p.pid in self.has_read_lock:
                                with self.read_lock:
                                    self.read_count.value -= 1
                                    self.has_read_lock.remove(p.pid)

                            if self.has_data_request_lock.value == p.pid:
                                self.has_data_request_lock.value = -1
                                try:
                                    self.data_request_lock.release()
                                except:
                                    pass

                            if self.has_write_lock.value == p.pid:
                                self.has_write_lock.value = -1
                                try:
                                    self.write_lock.release()
                                except:
                                    pass

                    for p in old_procs:
                        try:
                            self.procs.remove(p)
                        except ValueError:
                            logger.warn("Did not find %s in the list of processes.  Continuing on." % p.name)

                    for p in new_procs:
                        self.procs.append(p)
                        logger.warn("Started a new consumer (%s) to replace a zombie consumer" % p.name)
                        p.start()

                else:
                    # We got one.
                    retrieved += 1
                    if code is None:
                        logger.warn("Got an unrecognized response from a task.")
                    elif code == -1:
                        logger.warn("Particle %s has FAILED!!" % tempres.uid)
                    elif code == -2:
                        self.error_code = code
                        logger.warn("CachingDataController has FAILED!!  Removing cache file so the particles fail.")
                        try:
                            os.remove(self.cache_path)
                        except OSError:
                            logger.debug("Could not remove cache file, it probably never existed")
                            pass
                    elif code == -3:
                        self.error_code = code
                        logger.info("A zombie process was caught and task was removed from queue")
                    elif isinstance(tempres, Particle):
                        logger.info("Particle %d finished" % tempres.uid)
                        return_particles.append(tempres)
                        # We mulitply by 95 here to save 5% for the exporting
                        logger.progress((round((retrieved / self.total_task_count()) * 90., 1), "Particle %d finished" % tempres.uid))
                    elif tempres == "CachingDataController":
                        logger.info("CachingDataController finished")
                        logger.progress((round((retrieved / self.total_task_count()) * 90., 1), "CachingDataController finished"))
                    else:
                        logger.info("Got a strange result on results queue")
                        logger.info(str(tempres))

                    logger.info("Retrieved %i/%i results" % (int(retrieved), self.total_task_count()))

                # Relax
                time.sleep(1)

            if len(return_particles) != total_particles:
                logger.warn("Some particles failed and are not included in the output")

            # The results queue should be empty at this point
            assert self.results.empty() is True

            # Should be good to join on the tasks now that the queue is empty
            logger.info("Joining the task queue")
            self.tasks.join()
            self.tasks.close()
            self.tasks.join_thread()

        finally:
            # Join all processes
            logger.info("Joining the processes")
            for w in self.procs + [self.data_controller_process]:
                    # Wait 20 seconds
                    w.join(20.)
                    if w.is_alive():
                        # Process is hanging, kill it.
                        logger.info("Terminating %s forcefully.  This should have exited itself." % w.name)
                        w.terminate()

        if self.error_code == -2:
            raise ValueError("Error in the BaseDataController (error_code was -2)")

        results = ex.ResultsPyTable(output_h5_file)
        for p in return_particles:
            for x in range(len(p.locations)):
                results.write(p.timestep_index_dump(x))
        results.compute()
        results.close()

        return

    def cleanup(self):
        super(CachingModelController, self).cleanup()

        # Remove Manager so it shuts down
        del self.mgr

        # Remove the cache file
        if self.remove_cache is True:
            try:
                os.remove(self.cache_path)
            except OSError:
                logger.debug("Could not remove cache file, it probably never existed")



from datetime import timedelta
import numpy as np
import pytz


class SunCycles(object):

    SETTING = "sunset"
    RISING = "sunrise"

    @classmethod
    def cycles(cls, **kwargs):
        """
        Classmethod for convenience in returning both the sunrise and sunset
        based on a location and date.  Always calculates the sunrise and sunset on the
        given date, no matter the time passed into the function in the datetime object.

        Parameters:
            loc = Location4D (object)
                OR
            point = Shapely point (object)
            time = datetime in UTC (object)
                OR
            lat = latitude (float)
            lon = longitude (float)
            time = datetime in UTC (object)

        Returns:
            { 'sunrise': datetime in UTC, 'sunset': datetime in UTC }


        Sources:
            http://williams.best.vwh.net/sunrise_sunset_example.htm
        """
        if "loc" not in kwargs:
            if "point" not in kwargs:
                if "lat" not in kwargs or "lon" not in kwargs:
                    raise ValueError("You must supply some form of lat/lon coordinates")
                else:
                    lat = kwargs.get("lat")
                    lon = kwargs.get("lon")
            else:
                lat = kwargs.get("point").y
                lon = kwargs.get("point").x
            if "time" not in kwargs:
                raise ValueError("You must supply a datetime object")
            else:
                time = kwargs.get("time")
        else:
            lat = kwargs.get("loc").latitude
            lon = kwargs.get("loc").longitude
            time = kwargs.get("loc").time

        # Convert time to UTC.  Save passed in timezone to return later.
        if time.tzinfo is None:
            time = time.replace(tzinfo=pytz.utc)
            original_zone = pytz.utc
        else:
            original_zone = time.tzinfo

        local_jd = time.timetuple().tm_yday
        utc_jd = time.astimezone(pytz.utc).timetuple().tm_yday

        # We ALWAYS want to return the sunrise/sunset for the day that was passed
        # in (with timezone accounted for), regardless of what the UTC day is.  Modify
        # the UTC julian day here if need be.
        comp = (utc_jd > local_jd) - (utc_jd < local_jd)
        if comp == 1:
            utc_jd -= 1
        elif comp == -1:
            utc_jd += 1

        time = time.replace(hour=0, minute=0, second=0, microsecond=0)

        rising_h, rising_m = cls._calc(jd=utc_jd, lat=lat, lon=lon, stage=cls.RISING)
        setting_h, setting_m = cls._calc(jd=utc_jd, lat=lat, lon=lon, stage=cls.SETTING)

        # _calc returns UTC hours and minutes, so assume time is in UTC for a few lines...
        rising = time.replace(tzinfo=pytz.utc) + timedelta(hours=rising_h, minutes=rising_m)
        setting = time.replace(tzinfo=pytz.utc) + timedelta(hours=setting_h, minutes=setting_m)
        # LOOK: We may be adding 24 hours to the setting time.  Why?
        if setting < rising:
            setting = setting + timedelta(hours=24)

        rising = rising.astimezone(original_zone)
        setting = setting.astimezone(original_zone)

        return { cls.RISING : rising, cls.SETTING : setting}

    @classmethod
    def _calc(cls, **kwargs):
        """
        Calculate sunrise or sunset based on:
        Parameters:
            jd: Julian Day
            lat: latitude
            lon: longitude
            stage: sunrise or sunset
        """
        zenith = 90.833333 # official value

        jd = kwargs.get("jd", None)
        lat = kwargs.get("lat", None)
        lon = kwargs.get("lon", None)
        stage = kwargs.get("stage", None)
        if jd is None or stage is None or lat is None or lon is None:
            raise ValueError("Must supply an 'jd', 'lat, 'lon', and 'stage' parameter")

        if stage != SunCycles.RISING and stage != SunCycles.SETTING:
            raise ValueError("'stage' parameter must be %s or %s" % (SunCycles.RISING, SunCycles.SETTING))

        longhr = lon / 15.
        if stage == SunCycles.RISING:
            apx = jd + ( (6 - longhr) / 24 )
        elif stage == SunCycles.SETTING:
            apx = jd + ( (18 - longhr) / 24 )

        sun_mean_anom = ( 0.9856 * apx ) - 3.289 # sun's mean anomaly
        #sun's longitude
        sun_lon = sun_mean_anom + (1.916 * np.sin( np.radians(sun_mean_anom) )) \
                + (0.02 * np.sin( np.radians(2 * sun_mean_anom) )) + 282.634

        if sun_lon > 360:
            sun_lon = sun_lon - 360
        elif sun_lon < 0:
            sun_lon = sun_lon + 360

        right_ascension = np.degrees(np.arctan( 0.91764 * np.tan( np.radians(sun_lon) ) )) # sun's right ascension

        if right_ascension > 360:
            right_ascension = right_ascension - 360
        elif right_ascension < 0:
            right_ascension = right_ascension + 360

        # put sun's right ascension value in the same quadrant as the sun's
        # true longitude
        lQuad = 90. * np.floor(sun_lon / 90.)
        raQuad = 90. * np.floor(right_ascension / 90.)
        right_ascension = right_ascension + ( lQuad - raQuad)
        right_ascension = right_ascension / 15. # Convert to hours

        # Sun's declination
        sinDecl = 0.39782 * np.sin( np.radians(sun_lon) )
        cosDecl = np.cos( np.arcsin( sinDecl ) )

        # Sun's local hour angle
        cosHr = (np.cos( np.radians(zenith) ) - ( sinDecl * np.sin(np.radians(lat)) )) \
                / ( cosDecl * np.cos( np.radians(lat) ) )

        if cosHr > 1: # Sun doesn't rise on this loc on this date
            return -1, -1
        elif cosHr < -1: # Sun doesn't set on this location on this date
            return -1, -1
        elif stage == SunCycles.RISING: # Sunrise
            hr = 360 - np.degrees(np.arccos(cosHr))
        elif stage == SunCycles.SETTING:  # Sunset
            hr = np.degrees(np.arccos(cosHr))

        hr = hr / 15. # Convert angle to hours

        localTime = hr + right_ascension - ( 0.06571 * apx ) - 6.622# local meantime of rise/set

        UTtime = localTime - longhr # adjust to UTC

        if UTtime < 0:
            UTtime = UTtime + 24
        elif UTtime > 24:
            UTtime = UTtime - 24

        hour = np.floor(UTtime)

        minute = (UTtime - hour) * 60
        if minute == 60:
            hour = hour + 1
            minute = 0

        return hour, minute



import sys, os, bisect, netCDF4, math
import numpy as np
try:
    from osgeo import gdal
    from osgeo import osr
except ImportError:
    import osr
    import gdal

def compute_probability(trajectory_files, bbox=None,
                        nx=None, ny=None, method='overall', parameter='location'):
    """
        This function creates a probability (stochastic) grid
        for trajectory model data using 'overall' method (based
        on normalization by nsteps * nparticles) or 'run' method
        (based on normalization by run).

        probability_grid = compute_probability([myfile1.nc, myfile2.nc],
                                               bbox = [-75, 23, -60, 45],
                                               nx = 1000, ny = 1000,
                                               method = 'overall')
        stoch.compute_probability(['/media/sf_Python/trajectories.nc'],bbox=[-148,60,-146,61],nx=500,ny=500,method='overall',parameter='settlement')
    """
    xarray = np.linspace(float(bbox[0]), float(bbox[2]), int(nx)+1)
    yarray = np.linspace(float(bbox[1]), float(bbox[3]), int(ny)+1)
    if method=='overall':
        prob = np.zeros((ny, nx))
        for runfile in trajectory_files:
            run = netCDF4.Dataset(runfile)
            if parameter == 'location':
                lat = run.variables['lat'][:].flatten()
                lon = run.variables['lon'][:].flatten()
                column_i, row_i = [], []
                for clon, clat in zip(lon, lat):
                    column_i.append(bisect.bisect(xarray, clon))
                    row_i.append(bisect.bisect(yarray, clat))
                    try:
                        prob[row_i[-1], column_i[-1]] += 1
                    except Exception:
                        pass
            elif parameter == 'settlement':
                for i in range(run.variables['time'].shape[0]):
                    settle_index = np.where(run.variables['settled'][i,:]==1)
                    if len(settle_index[0]) > 0:
                        lat = run.variables['lat'][i, settle_index[0]].flatten()
                        lon = run.variables['lon'][i, settle_index[0]].flatten()
                        column_i, row_i = [], []
                        for clon, clat in zip(lon, lat):
                            column_i.append(bisect.bisect(xarray, clon))
                            row_i.append(bisect.bisect(yarray, clat))
                            try:
                                prob[row_i[-1], column_i[-1]] += 1
                            except Exception:
                                pass
            else:
                raise ValueError("Parameter for stochastic assessment not valid")
        shape = run.variables['time'].shape
        prob = prob / (shape[0] * len(trajectory_files)) # Assumes same # of particles
                                                         # for every run, may be bad
                                                         # assumtion
    elif method=='run':
        prob = []
        for i, runfile in enumerate(trajectory_files):
            prob.append(np.zeros((ny, nx)))
            run = netCDF4.Dataset(runfile)
            if parameter == 'location':
		        lat = run.variables['lat'][:].flatten()
		        lon = run.variables['lon'][:].flatten()
            elif parameter == 'settlement':
                settle_index = np.where(run.variables['settled'][-1,:]==1)[0]
                lat = run.variables['lat'][-1, settle_index].flatten()
                lon = run.variables['lon'][-1, settle_index].flatten()
            else:
                raise ValueError("Parameter for stochastic assessment not valid")
            column_i, row_i = [], []
            for clon, clat in zip(lon, lat):
                column_i.append(bisect.bisect(xarray, clon))
                row_i.append(bisect.bisect(yarray, clat))
                try:
	                if prob[i][row_i[-1], column_i[-1]] == 0:
	                	prob[i][row_i[-1], column_i[-1]] = 1
                except Exception:
                    pass
        prob2 = np.zeros((ny, nx))
        for run in prob:
            prob2 = run + prob2
        prob = prob2 / len(prob)
    return prob

def compute_probability_settle(trajectory_files, bbox=None,
                        nx=1000, ny=1000, method='overall'):
    """
        This function creates a probability (stochastic) grid
        for trajectory model data based on settlement location,
        normalized by run.

        probability_grid = compute_probability_settle([myfile1.nc, myfile2.nc],
                                               bbox = [-75, 23, -60, 45],
                                               nx = 1000, ny = 1000,
                                               method='overall')
    """
    prob = compute_probability(trajectory_files,
                               bbox,
                               nx, ny,
                               method,
                               parameter='settlement',
                              )
    return prob

def count_settlement(trajectory_file, bbox=None,
                        nx=1000, ny=1000):
    xarray = np.linspace(float(bbox[0]), float(bbox[2]), int(nx)+1)
    yarray = np.linspace(float(bbox[1]), float(bbox[3]), int(ny)+1)
    prob = np.zeros((ny, nx))
    run = netCDF4.Dataset(trajectory_file)
    #for i in range(run.variables['time'].shape[0]):
    settle_index = np.where(run.variables['settled'][-1,:]==1)
    if len(settle_index[0]) > 0:
        lat = run.variables['lat'][-1, settle_index[0]].flatten()
        lon = run.variables['lon'][-1, settle_index[0]].flatten()
        column_i = [bisect.bisect(xarray, clon)-1 for clon in lon]
        row_i = [bisect.bisect(yarray, clat)-1 for clat in lat]
        for row, col in zip(row_i, column_i):
            try:
                if row<0 or col<0:
                    pass
                else:
                    prob[row, col] += 1
            except Exception:
                pass
    return prob

def export_probability(outputname, **kwargs):
    """
        Calculate probability and export to gis raster/grid
        format.

        export_probability(prob_out,
                           trajectory_files = [myfiles1.nc, myfiles2.nc],
                           bbox = [-75, 23, -60, 45],
                           nx = 1000, ny = 1000,
                           method = 'overall')
    """
    bbox = kwargs.get('bbox', None)
    nx, ny = kwargs.get('nx', None), kwargs.get('ny', None)
    if bbox == None:
        raise ValueError('Must supply bbox keyword argument.')
    if nx == None or ny == None:
        raise ValueError('Must supply nx and ny keyword arguments.')
    prob = compute_probability(**kwargs)
    export_grid(outputname, prob, bbox, nx, ny)

def export_grid(outputname, grid, bbox, nx, ny, ):
    xres = (float(bbox[2]) - float(bbox[0])) / float(nx)
    yres = (float(bbox[3]) - float(bbox[1])) / float(ny)
    tiff = gdal.GetDriverByName('GTiff')
    rasterout = tiff.Create(outputname + '.tif', nx,
                            ny, 1, gdal.GDT_Float32)
    raster_xfrm = [float(bbox[0]), xres, 0.0, float(bbox[3]), 0.0, -yres]
    rasterout.SetGeoTransform(raster_xfrm)
    srs = osr.SpatialReference()
    srs.SetWellKnownGeogCS( 'WGS84' )
    rasterout.SetProjection( srs.ExportToWkt() )
    rasterout.GetRasterBand(1).WriteArray(grid[::-1, :])




import math

from paegan.utils.asamath import AsaMath
from paegan.utils.asarandom import AsaRandom
from paegan.transport.utils.asatransport import AsaTransport
from paegan.transport.models.base_model import BaseModel

from paegan.logger import logger

class Transport(BaseModel):
    """
        Transport a particle in the x y and w direction. Requires horizontal and vertical dispersion coefficients.
        Will only move particle when self.move() is called with the proper arguments.
    """

    def __init__(self, **kwargs):

        if "horizDisp" and "vertDisp" in kwargs:
            self._horizDisp = float(kwargs.pop('horizDisp'))
            self._vertDisp = float(kwargs.pop('vertDisp'))
        else:
            raise TypeError( "must provide a horizontal and vertical dispersion coefficient (horizDisp and vertDisp)" )

    def set_horizDisp(self, hdisp):
        self._horizDisp = hdisp
    def get_horizDisp(self):
        return self._horizDisp
    horizDisp = property(get_horizDisp, set_horizDisp)

    def set_vertDisp(self, vdisp):
        self._vertDisp = vdisp
    def get_vertDisp(self):
        return self._vertDisp
    vertDisp = property(get_vertDisp, set_vertDisp)

    def move(self, particle, u, v, w, modelTimestep, **kwargs):
        """
        Returns the lat, lon, H, and velocity of a projected point given a starting
        lat and lon (dec deg), a depth (m) below sea surface (positive up), u, v, and w velocity components (m/s), a horizontal and vertical
        displacement coefficient (m^2/s) H (m), and a model timestep (s).

        GreatCircle calculations are done based on the Vincenty Direct method.

        Returns a dict like:
            {   'latitude': x, 
                'azimuth': x,
                'reverse_azimuth': x, 
                'longitude': x, 
                'depth': x, 
                'u': x
                'v': x, 
                'w': x, 
                'distance': x, 
                'angle': x, 
                'vertical_distance': x, 
                'vertical_angle': x }
        """

        logger.debug("U: %s, V: %s, W: %s" % (str(u),str(v),str(w)))

        # IMPORTANT:
        # If we got no data from the model, we are using the last available value stored in the particles!
        if (u is None) or (u is not None and math.isnan(u)):
            u = particle.last_u()
        if (v is None) or (v is not None and math.isnan(v)):
            v = particle.last_v()
        if (w is None) or (w is not None and math.isnan(w)):
            w = particle.last_w()

        particle.u_vector = u
        particle.v_vector = v
        particle.w_vector = w

        if particle.halted:
            u,v,w = 0,0,0
        else:
            u += AsaRandom.random() * ((2 * self._horizDisp / modelTimestep) ** 0.5) # u transformation calcualtions
            v += AsaRandom.random() * ((2 * self._horizDisp / modelTimestep) ** 0.5) # v transformation calcualtions
            w += AsaRandom.random() * ((2 * self._vertDisp / modelTimestep) ** 0.5) # w transformation calculations

        result = AsaTransport.distance_from_location_using_u_v_w(u=u, v=v, w=w, timestep=modelTimestep, location=particle.location)
        result['u'] = u
        result['v'] = v
        result['w'] = w
        return result

    def __str__(self):
        return  """
            ***** Transport *****
              horizDisp: %s
              vertDisp:  %s
            """ % (str(self.horizDisp), str(self.vertDisp)) 
